// -*- C++ -*-

#pragma once
#include "gpu_defs.h"
#include "gpu_types.h"
#include "proto.h"


namespace goopax {
namespace impl {

namespace machinery {
template<typename T, typename = void>
struct pas;
}

struct tex_arg_base;
struct kernel_impl;
struct full_kernel;
class rundata;

namespace ns_param_kernel {
template<class R, class T>
struct pas;
template<class R, class T, class REST, bool is_gpu, bool is_resource, bool is_gather>
struct pas_split;
}
template<typename T, typename = void>
struct pas;

namespace interface {
extern "C" {

GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_future_get(rundata* rd, size_t*, const paramtype_t**) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) errno_type goopax_41E_future_get(rundata* rd, unsigned int id, void* dest) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_rd_finished(bool*, rundata*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) size_t goopax_4_rd_get_info(const rundata*, int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_rd_release(rundata* rd) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) void goopax_4_future_set_callback(rundata* rd, std::function<void()>* funccopy, void (&caller)(std::function<void()>*), void (&deleter)(std::function<void()>*)) noexcept;

static void callback_caller(std::function<void()>* funcp) noexcept
{
    (*funcp)();
}
static void callback_deleter(std::function<void()>* funcp) noexcept
{
    delete funcp;
}
}
}

struct goopax_future_base
{
    rundata*     rd2        = nullptr;
    mutable bool did_finish = false;
    unsigned int id;
    void         wait_common(size_t& nt, const paramtype_t*& p) const
    {
        if (rd2 == nullptr)
        {
            exwhat() << "Cannot wait for kernel that hasn't even started yet.";
            throw EX::syntax();
        }
        check_errno(interface::goopax_4E_future_get(rd2, &nt, &p));
        did_finish = true;
    }
    void wait_common(void* dest) const
    {
        if (rd2 == nullptr)
        {
            exwhat() << "Cannot wait for kernel that hasn't even started yet.";
            throw EX::syntax();
        }
        check_errno(interface::goopax_41E_future_get(rd2, id, dest));
        did_finish = true;
    }

    goopax_future_base()                          = default;
    goopax_future_base(const goopax_future_base&) = delete;
    goopax_future_base& operator=(const goopax_future_base&) = delete;
    inline goopax_future_base(goopax_future_base&& b);
    inline goopax_future_base& operator=(goopax_future_base&& b);
    inline ~goopax_future_base();
};


/// \ingroup kernel
/// Common future class base, provides information about the kernel execution.
class goopax_future_info : protected goopax_future_base
{
    template<typename T, typename SIZE_TYPE>
    friend class buffer;
    friend class rundata;

    public:
    /// returns true if kernel has finished execution
    bool finished() const
    {
        if (!did_finish)
        {
            return if_call(interface::goopax_4E_rd_finished, rd2);
        }
        return did_finish;
    }

    /// The provided function will be called when the operation has finished.
    /// The provided function must return swiftly and it must not call any goopax functions.
    void set_callback(std::function<void()> func)
    {
        if (did_finish)
        {
            func();
        }
        else
        {
            auto* funccopy = new std::function<void()>(func);
            interface::goopax_4_future_set_callback(rd2, funccopy, interface::callback_caller, interface::callback_deleter);
        }
    }

    /// number of threads per workgroup
    inline size_t local_size() const
    {
        if (rd2 == nullptr)
        {
            exwhat() << "Cannot get thread info for kernel that hasn't started yet.";
            throw EX::syntax();
        }
        return interface::goopax_4_rd_get_info(this->rd2, 0);
    }
    /// total number of threads
    inline size_t global_size() const
    {
        if (rd2 == nullptr)
        {
            exwhat() << "Cannot get thread info for kernel that hasn't started yet.";
            throw EX::syntax();
        }
        return interface::goopax_4_rd_get_info(this->rd2, 1);
    }
    /// number of workgroups
    inline size_t num_groups() const
    {
        if (rd2 == nullptr)
        {
            exwhat() << "Cannot get thread info for kernel that hasn't started yet.";
            throw EX::syntax();
        }
        return interface::goopax_4_rd_get_info(this->rd2, 2);
    }
};


/// \ingroup kernel
/// future class for gather return values
/// \tparam T return type. May be a C++ intrinsic type, or a goopax struct type
template<typename T>
class goopax_future : public goopax_future_info
{
    template<class R, class U>
    friend struct ns_param_kernel::pas;
    template<class R, class U>
    friend struct machinery::pas;
    template<class R, class U, class REST, bool is_gpu, bool is_resource, bool is_gather>
    friend struct ns_param_kernel::pas_split;

    public:
    using value_type = T;

    private:
    mutable T    data;
    unsigned int offset;
    T(*collect)
    (const paramtype_t*, size_t); 

    void set(rundata& rd0, unsigned int offset0, T (*collect0)(const paramtype_t*, size_t))
    {
        this->rd2     = &rd0;
        this->offset  = offset0;
        this->collect = collect0;
    }

    public:
    /// wait for kernel to finish
    void wait() const
    {
        if (!did_finish)
        {
            size_t             nt;
            const paramtype_t* p;
            if (collect != nullptr)
            {
                wait_common(nt, p);
                data = collect(p + offset * nt, nt);
            }
            else
            {
                wait_common(&data);
            }
        }
    }
    /// wait for kernel to finish and get value
    T get() const
    {
        /// \return value gathered from all threads and combined using the gather operation
        wait();
        return data;
    }

    /// default constructor
    goopax_future() = default;

    goopax_future(const goopax_future&) = delete;
    goopax_future& operator=(const goopax_future&) = delete;

    /// move semantics
    goopax_future(goopax_future&&) = default;
    /// move semantics
    goopax_future& operator=(goopax_future&&) = default;
};

/// \ingroup kernel
/// future class, Returned from kernels that don't return a value.
/// \tparam T return type. May be a C++ intrinsic type, or a goopax struct type
template<>
class goopax_future<void> : public goopax_future_info
{
    template<class R, class U>
    friend struct ns_param_kernel::pas;
    template<typename T, typename>
    friend struct machinery::pas;
    template<class R, class U, class REST, bool is_gpu, bool is_resource, bool is_gather>
    friend struct ns_param_kernel::pas_split;
    template<typename T, typename SIZE_TYPE>
    friend class buffer;
    template<unsigned int DIM, typename T, bool normalized>
    friend class image_buffer;
    template<unsigned int DIM, typename T, bool normalized>
    friend class image_array_buffer;

    void set(rundata& rd0)
    {
        this->rd2 = &rd0;
    }

    public:
    /// wait for kernel to finish
    void wait() const
    {
        if (!did_finish)
        {
            size_t             nt;
            const paramtype_t* p;
            wait_common(nt, p);
        }
    }
    /// default constructor
    goopax_future() = default;

    goopax_future(const goopax_future&) = delete;
    goopax_future& operator=(const goopax_future&) = delete;

    /// move semantics
    goopax_future(goopax_future&&) = default;
    /// move semantics
    goopax_future& operator=(goopax_future&&) = default;
};


goopax_future_base::goopax_future_base(goopax_future_base&& b)
    : rd2(b.rd2)
    , did_finish(b.did_finish)
    , id(b.id)
{
    b.rd2 = nullptr;
}

goopax_future_base& goopax_future_base::operator=(goopax_future_base&& b)
{
    if (rd2 != nullptr)
    {
        interface::goopax_4_rd_release(rd2);
    }
    rd2        = b.rd2;
    b.rd2      = nullptr;
    did_finish = b.did_finish;
    id         = b.id;
    return *this;
}

goopax_future_base::~goopax_future_base()
{
    if (rd2 != nullptr)
    {
        interface::goopax_4_rd_release(rd2);
    }
}


}
}
