#pragma once

#include "kernel.h"

namespace goopax {
namespace impl {

namespace machinery {
template<typename OP, typename T>
void collectfunc(const paramtype_t* p, unsigned int threads, void* dest)
{
    static_assert(!is_gpu_type<T>::value);
    T ret = pas_data::read<T>(p, 0);
    for (unsigned int k = 1; k < threads; ++k)
    {
        ret = OP::op(ret, pas_data::read<T>(p, k * pas_data::paramsize<T>()));
    }
    *static_cast<T*>(dest) = ret;
}
}


template<typename T>
struct gather_result
{
    using gpu_T = typename make_gpu<T>::type;
    std::function<gpu_T()> reducefunc;
    void (*collectfunc)(const paramtype_t*, unsigned int, void*) = nullptr;

    gather_result()
    {}

    template<typename G>
    gather_result(const G& gather)
        : reducefunc([x = static_cast<gpu_T>(gather)]() {
            return work_group_reduce<G>(x, gpu_thread::local_size());
        })
        , collectfunc(machinery::collectfunc<G, T>)
    {}
};

#if __cplusplus >= 201700
template<typename OP>
gather_result(OP) -> gather_result<typename gettype<typename OP::value_type>::cpu>;
#endif


template<typename... ARGS>
struct add;

template<>
struct add<> : public integral_constant<size_t, 0>
{};

template<typename T, typename... ARGS>
struct add<T, ARGS...> : public integral_constant<size_t, T::value + add<ARGS...>::value>
{};


namespace machinery {

template<typename K, typename ARGS>
struct paramloop;

template<typename T>
struct pas<T, typename enable_if<is_arithmetic<T>::value || is_gpu_struct<T>::value>::type>
{
    using gpu_T        = typename gettype<T>::gpu;
    using param_size   = integral_constant<size_t, pas_data::paramsize<T>()>;
    using gather_size  = integral_constant<size_t, 0>;
    using gather_num   = integral_constant<size_t, 0>;
    using num_tex_args = integral_constant<size_t, 0>;
    template<typename REST, typename K, typename... P>
    static auto impl(K& kernel, unsigned int& offset, unsigned int& parampos, P&... p)
    {
        const auto data = pas_data::read<gpu_T>(parampos);
        parampos += param_size::value;
        return paramloop<K, REST>::impl(kernel, offset, parampos, p..., data);
    }
    template<typename PARAM, typename V>
    static inline void prepare(PARAM& param, V&, const T& b)
    {
        pas_data::write(param, 0, b);
        param += param_size::value;
    }
    static inline void setcollect(rundata&, unsigned int&, unsigned int&, const T&) {}
};


template<typename T>
struct pas<goopax_future<T>&>
{
    using param_size   = integral_constant<size_t, 0>;
    using gather_size  = integral_constant<size_t, pas_data::paramsize<T>()>;
    using gather_num   = integral_constant<size_t, 1>;
    using num_tex_args = integral_constant<size_t, 0>;
    using gpu_T        = typename gettype<T>::gpu;

    template<typename REST, typename K, typename... P>
    static auto impl(K& kernel, unsigned int& offset, unsigned int& parampos, P&... p)
    {
        gather_result<T>   result;
        const unsigned int current_offset = offset;
        offset += gather_size::value;

        if constexpr (is_same<decltype(paramloop<K, REST>::impl(kernel, offset, parampos, p..., result)), void>::value)
        {
            paramloop<K, REST>::impl(kernel, offset, parampos, p..., result);
            interface::goopax_4_fk_add_collect(current_offset, *result.collectfunc);

            const gpu_T tot = result.reducefunc();
            gpu_if(local_id() == 0)
            {
                pas_data::write_gather(current_offset * num_groups() + group_id() * pas_data::paramsize<T>(), tot);
            }
        }
        else
        {
            auto ret = paramloop<K, REST>::impl(kernel, offset, parampos, p..., result);
            interface::goopax_4_fk_add_collect(current_offset, *result.collectfunc);

            const gpu_T tot = result.reducefunc();
            gpu_if(local_id() == 0)
            {
                pas_data::write_gather(current_offset * num_groups() + group_id() * pas_data::paramsize<T>(), tot);
            }
            return ret;
        }
    }

    template<typename PARAM, typename V>
    static inline void prepare(PARAM&, V&, goopax_future<T>&)
    {
    }

    static inline void setcollect(rundata& rd, unsigned int& offset, unsigned int& id, goopax_future<T>& b)
    {
        b = goopax_future<T>();
        b.set(rd, offset, nullptr);
        b.id = id++;
        offset += gather_size::value;
    }
};

template<>
struct pas<goopax_future<void>&>
{
    using param_size  = integral_constant<size_t, 0>;
    using gather_size = integral_constant<size_t, 0>;
    using gather_num  = integral_constant<size_t, 1>;

    template<typename REST, typename K, typename... P>
    static auto impl(K& kernel, unsigned int& offset, unsigned int& parampos, P&... p)
    {
        return paramloop<K, REST>::impl(kernel, offset, parampos, p...);
    }

    template<typename PARAM, typename V>
    static inline void prepare(PARAM&, V&, goopax_future<void>&)
    {
    }

    static inline void setcollect(rundata& rd, unsigned int&, unsigned int&, goopax_future<void>& b)
    {
        b.set(rd);
    }
};

template<typename BUF>
struct pas_buf
{
    using param_size   = integral_constant<size_t, 0>;
    using gather_size  = integral_constant<size_t, 0>;
    using gather_num   = integral_constant<size_t, 0>;
    using num_tex_args = integral_constant<size_t, 1>;

    template<typename REST, typename K, typename... P>
    static auto impl(K& kernel, unsigned int& offset, unsigned int& parampos, P&... p)
    {
        typename remove_reference<BUF>::type::resource_type res;
        return paramloop<K, REST>::impl(kernel, offset, parampos, p..., res);
    }
    template<typename PARAM, typename V>
    static inline void prepare(PARAM&, V& vp, BUF b)
    {
        *vp++ = b.get_buf();
    }

    static inline void setcollect(rundata&, unsigned int&, unsigned int&, BUF) {}
};

template<typename T, typename SIZE>
struct pas<buffer<T, SIZE>&> : pas_buf<buffer<T, SIZE>&>
{};

template<typename T, typename SIZE>
struct pas<const buffer<T, SIZE>&> : pas_buf<const buffer<T, SIZE>&>
{};


template<unsigned int DIM, typename T, bool normalized>
struct pas<image_buffer<DIM, T, normalized>&> : pas_buf<image_buffer<DIM, T, normalized>&>
{};

template<unsigned int DIM, typename T, bool normalized>
struct pas<const image_buffer<DIM, T, normalized>&> : pas_buf<const image_buffer<DIM, T, normalized>&>
{};


template<unsigned int DIM, typename T, bool normalized>
struct pas<image_array_buffer<DIM, T, normalized>&> : pas_buf<image_array_buffer<DIM, T, normalized>&>
{};

template<unsigned int DIM, typename T, bool normalized>
struct pas<const image_array_buffer<DIM, T, normalized>&> : pas_buf<const image_array_buffer<DIM, T, normalized>&>
{};


template<typename K, typename T, typename... A>
struct paramloop<K, tuple<T, A...>>
{
    template<typename... P>
    static auto impl(K& k, unsigned int& offset, unsigned int& parampos, P&... p)
    {
        return pas<T>::template impl<tuple<A...>>(k, offset, parampos, p...);
    }

    template<typename PARAM, typename V>
    static inline void prepare(PARAM& param, V& vp, T& t, A... p)
    {
        pas<T>::prepare(param, vp, t);
        paramloop<K, tuple<A...>>::prepare(param, vp, p...);
    }

    static inline void setcollect(rundata& rd, unsigned int& offset, unsigned int& id, T& t, A... p)
    {
        pas<T>::setcollect(rd, offset, id, t);
        paramloop<K, tuple<A...>>::setcollect(rd, offset, id, p...);
    }
};


template<typename K>
struct paramloop<K, tuple<>>
{
    template<typename... P>
    static auto impl(K& kernel, unsigned int&, unsigned int&, P&... p)
    {
        return kernel.codefunc(p...);
    }
    template<typename PARAM, typename V>
    static inline void prepare(PARAM&, V&)
    {}

    static inline void setcollect(rundata&, unsigned int&, unsigned int&) {}
};
}


/// \ingroup new kernel
/// \brief kernel class to derive gpu kernels from
/// \param K derived kernel class.
/// If The C++ language standard is < 14, any return type of the kernel must be specified with the RETURN_TYPE template argument.
/// All gpu kernels must be derived from this class, where the new class must be passed as template argument, like so:
/// struct my_kernel :
/// public kernel<my_kernel>
/// {
///    void program(...) {...}
/// };

class kernel_base : public full_kernel
{
    public:
    goopax_device device;

    /// \name overriding default thread numbers
    ///\{

    /// set the local size for this kernel.
    inline void force_local_size(int local_size) { interface::goopax_4_fk_force_set(impl_p, device.get_impl(), 0, local_size); }
    /// set the global size for this kernel.
    inline void force_global_size(int global_size) { interface::goopax_4_fk_force_set(impl_p, device.get_impl(), 1, global_size); }
    /// \}

    inline void force_compile() const { check_errno(interface::goopax_4E_fk_force_compile(impl_p, device.get_impl())); }
};

template<typename RET, typename... ARGS>
class kernel<RET(ARGS...)> : public kernel_base
{
    template<typename K, typename A>
    friend struct machinery::paramloop;

    public:
    using return_type      = goopax_future<RET>;
    using code_return_type = typename conditional<is_same<RET, void>::value, void, return_type>::type;
    using code_t           = typename make_gpu<code_return_type(ARGS...)>::type;


    private:
    using param_size   = add<typename machinery::pas<return_type&>::param_size, typename machinery::pas<ARGS>::param_size...>;
    using gather_size  = add<typename machinery::pas<return_type&>::gather_size, typename machinery::pas<ARGS>::gather_size...>;
    using gather_num   = add<typename machinery::pas<return_type&>::gather_num, typename machinery::pas<ARGS>::gather_num...>;
    using num_tex_args = add<typename machinery::pas<ARGS>::num_tex_args...>;

    std::function<code_t> codefunc;

    virtual void implcall(unsigned int* p) final override
    {
        //        using functype = decltype(&K::program);
        *p = param_size::value;

        unsigned int parampos = 0;
        if constexpr (is_same<RET, void>::value)
        {
            unsigned int offset = 0;
            machinery::paramloop<kernel, tuple<ARGS...>>::impl(*this, offset, parampos);
        }
        else
        {
            using T                 = typename return_type::value_type;
            using gpu_T             = typename gettype<T>::gpu;
            unsigned int     offset = machinery::pas<return_type&>::gather_size::value;
            gather_result<T> result = machinery::paramloop<kernel, tuple<ARGS...>>::impl(*this, offset, parampos);
            interface::goopax_4_fk_add_collect(0, *result.collectfunc);
            const gpu_T tot = result.reducefunc();
            gpu_if(local_id() == 0)
            {
                pas_data::write_gather(group_id() * pas_data::paramsize<T>(), tot);
            }
        }
    }

    public:
    /// Operator (). This starts the gpu kernel on the device. The parameters must match the arguments specified in the function 'program()', exept that cpu types are passed, where gpu types are specified in the function 'program()'. The device may be specified as a first argument. This is optional except if the device cannot be deduced from the other arguments. All buffers must reside on the same device, which is the device on which the kernel is executed.
    //    template<class... A>
    inline return_type operator()(ARGS... p) const
    {
        //        using functype = decltype(&K::program);
        //        using ct       = change_type<0, tuple<>, typename fi::params>;
        //        using ct1      = change_type<1, tuple<>, typename fi::params>;
        //        using kp       = kernelparams<RET(ARGS...), typename ct::type, typename ct1::type>;
        //        using kp = kernelparams<RET(ARGS...), tuple<typename make_debug_wrapper<ARGS>::type...>, typename fi::params>;
        return_type ret;

        array<paramtype_t, param_size::value> pd;
        auto                                  pdp = pd.begin();
        array<texbuf*, num_tex_args::value>   vec;
        auto                                  vp = vec.begin();
        machinery::paramloop<kernel, tuple<return_type&, ARGS...>>::prepare(pdp, vp, ret, p...);
        assert(vp == vec.end());
        assert(pdp == pd.end());

        rundata*     rd     = if_call(interface::goopax_4E_fk_arun, device.get_impl(), this->impl_p, vec.data(), (unsigned int)vec.size(), pd.data(), gather_num::value);
        unsigned int offset = 0;
        unsigned int id     = 0;

        machinery::paramloop<kernel, tuple<return_type&, ARGS...>>::setcollect(*rd, offset, id, ret, p...);
        return ret;
    }

    void assign(goopax_device         device,
                std::function<code_t> code)
    {
        *this = kernel(device, code);
    }

    unsigned int local_size() const
    {
        unsigned int ret;
        check_errno(interface::goopax_4E_fk_get(this->impl_p, device.get_impl(), 0, &ret));
        return ret;
    }

    unsigned int global_size() const
    {
        unsigned int ret;
        check_errno(interface::goopax_4E_fk_get(this->impl_p, device.get_impl(), 1, &ret));
        return ret;
    }

    unsigned int num_groups() const
    {
        return global_size() / local_size();
    }

    /// constructor
    kernel(goopax_device         device,
           std::function<code_t> code)
        : codefunc(code)
    {
        this->device = device;

        const string name = pretty_typename(typeid(RET(ARGS...)));
        this->impl_p      = if_call(interface::goopax_4E_fk_new,
                               static_cast<full_kernel*>(this),
                               GOOPAX_CERT, sig,
                               fk_call_impl<>,
                               name.c_str(),
                               param_size::value,
                               gather_size::value);
    }

    kernel(const kernel& b)
        : kernel_base(b)
        , codefunc(b.codefunc)
    {
        full_kernel::inc();
    }

    kernel() = default;

    kernel& operator=(const kernel& b)
    {
        kernel_base::operator=(b);
        codefunc             = b.codefunc;
        full_kernel::inc();
        return *this;
    }
    ~kernel()
    {
        if (this->impl_p != nullptr)
        {
            device.wait_all();
        }
        full_kernel::dec();
    }
};


namespace machinery {
template<typename T>
struct getfunc;

template<typename C, typename R, typename... A>
struct getfunc<R (C::*)(A...)>
{
    using type = R(A...);
};

template<typename C, typename R, typename... A>
struct getfunc<R (C::*)(A...) const>
{
    using type = R(A...);
};

template<typename FUNC>
struct getfunc
{
    using type = typename getfunc<decltype(&FUNC::operator())>::type;
};


template<typename T>
struct remove_future
{
    using type = T;
};
template<typename T>
struct remove_future<goopax_future<T>>
{
    using type = T;
};
template<typename R, typename... A>
struct remove_future<R(A...)>
{
    using type = typename remove_future<R>::type(A...);
};


}

#if __cplusplus >= 201700
template<typename FUNC>
kernel(const goopax_device& device, const FUNC& func) -> kernel<typename machinery::remove_future<typename make_cpu<typename machinery::getfunc<FUNC>::type>::type>::type>;
#endif

}
using impl::gather_result;
using impl::kernel;
using impl::kernel_base;

}
