// -*- C++ -*-

#pragma once

#include "device.h"
#include "gpu_types.h"

struct _cl_mem;
struct _cl_device_id;


namespace cl {
class Event;
class CommandQueue;
}


namespace goopax {
/// \ingroup image
/// Image access flags
enum IMAGE_FLAGS
{
    IMAGE_READ_WRITE    = 1, ///< read/write access
    IMAGE_READ_ONLY     = 2, ///< read-only access
    IMAGE_WRITE_ONLY    = 4, ///< write-only access
    IMAGE_WRITE_DISCARD = 8  ///< write-only, previous values are not preserved
};

/// \ingroup memory
/// Buffer access flags
enum BUFFER_FLAGS
{
    BUFFER_READ_WRITE    = 1, ///< read/write access
    BUFFER_READ_ONLY     = 2, ///< read-only access
    BUFFER_WRITE_ONLY    = 4, ///< write-only access
    BUFFER_WRITE_DISCARD = 8  ///< write-only, previous values are not preserved
};

inline IMAGE_FLAGS operator|(IMAGE_FLAGS a, IMAGE_FLAGS b)
{
    return (IMAGE_FLAGS)((int)a | (int)b);
}
inline BUFFER_FLAGS operator|(BUFFER_FLAGS a, BUFFER_FLAGS b)
{
    return (BUFFER_FLAGS)((int)a | (int)b);
}


namespace impl {
struct texbuf;
namespace gpu_struct {
template<typename STR>
struct type_probe;
template<typename STR, typename TO>
struct changetype;
}

namespace interface {
extern "C" {
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_utils_get_hash(texbuf& buf, unsigned int hash[8]) noexcept;
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_utils_get_hash_ignore_uninit(texbuf& buf, unsigned int hash[8]) noexcept;
}
}


template<typename T = void>
struct init_caller
{
    init_caller() {}
    template<class FUNC, typename... A>
    init_caller(FUNC func, A... a)
    {
        func(a...);
    }
};


struct cl_kernel;
struct texture_base;


namespace interface {
extern "C" {
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_tab_lock(texbuf*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_tab_unlock(texbuf*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) size_t goopax_4_tab_size(const texbuf*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) goopax_device_impl* goopax_4_tab_get_device(const texbuf*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_tex_lock_map(texbuf*, void**, size_t, size_t) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_tex_release_map(texbuf*) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_tex_create_re(texbuf**, texbuf*, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_tex_create(texbuf**, goopax_device_impl* device, size_t, BUFFER_FLAGS, unsigned int*, void*, unsigned int, const gpu_struct::mapframe_base*, const gpu_struct::mapframe_base*) noexcept;
#if GOOPAX_HAVE_OPENCL
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_tex_create_cl(texbuf**, goopax_device_impl* device, void (*)(texbuf&, texbuf&), _cl_mem*, unsigned int) noexcept;
#endif
#if GOOPAX_HAVE_METAL && defined(__OBJC__)
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_tex_create_metal(texbuf**, goopax_device_impl* device, ::id, unsigned int) noexcept;
#endif
#if GOOPAX_HAVE_CUDA
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_tex_create_cuda(texbuf**, goopax_device_impl* device, void (*)(texbuf&, texbuf&), void*, size_t, unsigned int) noexcept;
#endif

[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_copy(texbuf*, texbuf*, size_t, size_t, size_t) noexcept;
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_copy_to_host(texbuf*, void*, size_t, size_t) noexcept;
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_copy_from_host(texbuf*, const void*, size_t, size_t) noexcept;
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_fill_buffer(texbuf*, const void*, size_t, size_t, size_t) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) errno_type goopax_41E_copy(rundata**, texbuf*, texbuf*, size_t, size_t, size_t) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) errno_type goopax_41E_copy_to_host(rundata**, texbuf*, void*, size_t, size_t) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) errno_type goopax_41E_copy_from_host(rundata**, texbuf*, const void*, size_t, size_t) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) errno_type goopax_41E_fill_buffer(rundata**, texbuf*, const void*, size_t, size_t, size_t) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_resource_info(const resource_impl*, gpu_writer_base_impl*, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_get_ptr(resource_impl*, gpu_writer_base_impl*) noexcept;
}
}

struct map_base
{
    texbuf* buf = nullptr;

    map_base(const map_base&) = delete;
    map_base& operator=(const map_base&) = delete;

    map_base(map_base&& b)
    {
        swap(buf, b.buf);
    }
    map_base& operator=(map_base&& b)
    {
        if (buf != nullptr)
            check_errno(interface::goopax_4E_tex_release_map(buf));
        buf   = b.buf;
        b.buf = nullptr;
        return *this;
    }

    map_base() = default;
    ~map_base()
    {
        if (buf != nullptr)
            check_errno(interface::goopax_4E_tex_release_map(buf));
    }
};


struct tex_arg_base
{
    texbuf*              buf = nullptr;
    inline goopax_device get_device() const
    {
        if (buf == nullptr)
            return goopax_device();
        else
            return interface::goopax_4_tab_get_device(buf);
    }

    [[deprecated]] inline array<unsigned int, 8> get_hash() const
    {
        array<unsigned int, 8> ret;
        check_errno(interface::goopax_4E_utils_get_hash(*buf, &ret[0]));
        return ret;
    }
    [[deprecated]] inline array<unsigned int, 8> get_hash_ignore_uninit() const
    {
        array<unsigned int, 8> ret;
        check_errno(interface::goopax_4E_utils_get_hash_ignore_uninit(*buf, &ret[0]));
        return ret;
    }

    inline tex_arg_base(texbuf* buf0)
        : buf(buf0)
    {}

    inline tex_arg_base(tex_arg_base&& b)
        : buf(b.buf)
    {
        b.buf = nullptr;
    }
    inline tex_arg_base& operator=(tex_arg_base&& b)
    {
        interface::goopax_4_tab_unlock(buf);
        buf   = b.buf;
        b.buf = nullptr;
        return *this;
    }

    virtual ~tex_arg_base()
    {
        interface::goopax_4_tab_unlock(buf);
    }
    tex_arg_base()                    = default;
    tex_arg_base(const tex_arg_base&) = delete;
    tex_arg_base& operator=(const tex_arg_base&) = delete;
};


struct texture_base : public tex_arg_base
{
    inline size_t bytesize() const { return interface::goopax_4_tab_size(buf); }

    texture_base(texbuf* buf)
        : tex_arg_base(buf)
    {}
    texture_base() = default;
};


template<typename A, typename B>
struct combine_size_type
{
    using type = typename conditional<(gettype<typename A::size_type>::size < gettype<typename B::size_type>::size), typename B::size_type, typename A::size_type>::type;
};


template<class T, typename size_type>
struct uav;


template<class T, typename CPU_SIZE_TYPE>
struct texture : public texture_base
{
    using cpu_size_type  = CPU_SIZE_TYPE;
    using cpu_ssize_type = typename make_signed<cpu_size_type>::type;
    using gpu_size_type  = typename gettype<cpu_size_type>::gpu;
    using gpu_ssize_type = typename gettype<cpu_ssize_type>::gpu;
    using size_type      = cpu_size_type;
    using ssize_type     = cpu_ssize_type;

    using resource_type = uav<T, cpu_size_type>;

    inline texbuf* get_buf() const
    {
        return this->buf;
    }

    inline texture(goopax_device device, const size_t size, BUFFER_FLAGS flags, unsigned int* GLres, void* host_ptr, const gpu_struct::mapframe_base* mf, const gpu_struct::mapframe_base* mfu)
        : texture_base(if_call(interface::goopax_4E_tex_create, device.get_impl(), size, flags, GLres, host_ptr, type_enum<T>::value, mf, mfu))
    {}

    inline texture(texbuf& base, INITTEX_REINTERPRET)
        : texture_base(if_call(interface::goopax_4E_tex_create_re, &base, type_enum<T>::value))
    {}

#if GOOPAX_HAVE_OPENCL
    inline texture(goopax_device device, _cl_mem* mem)
        : texture_base(if_call(interface::goopax_4E_tex_create_cl, device.get_impl(), nullptr, mem, type_enum<T>::value))
    {}
#endif
#if GOOPAX_HAVE_METAL && defined(__OBJC__)
    inline texture(goopax_device device, ::id mem)
        : texture_base(if_call(interface::goopax_4E_tex_create_metal, device.get_impl(), mem, type_enum<T>::value))
    {}
#endif
#if GOOPAX_HAVE_CUDA
    inline texture(goopax_device device, void* mem, size_t size)
        : texture_base(if_call(interface::goopax_4E_tex_create_cuda, device.get_impl(), nullptr, mem, size, type_enum<T>::value))
    {}
#endif


    texture()               = default;
    texture(const texture&) = delete;
    texture& operator=(const texture&) = delete;
    inline texture(texture&&)          = default;
    inline texture& operator=(texture&&) = default;
};

}
using impl::tex_arg_base;
}
