// -*- C++ -*-

#pragma once

#include "gpu_types.h"
#include "struct.h"

namespace goopax {
namespace impl {


template<typename T, memory::scope scope>
class gpu_type<T*, scope> : public wrapper<gpu_type<T*, scope>, gpu_writer_base_impl>
{
    private:
    using TT = typename remove_cv<T>::type;

    public:
    using size_type       = typename conditional<scope == memory::device, uint64_t, uint32_t>::type;
    using ssize_type      = typename make_signed<size_type>::type;
    using gpu_size_type   = typename gettype<size_type>::gpu;
    using gpu_ssize_type  = typename gettype<ssize_type>::gpu;
    using value_type      = typename goopax_struct_changetype<TT, xtype_gpu>::type;
    using basetype        = typename buffer<TT>::base_value_type;
    using reference       = value_type&;
    using const_reference = const value_type&;
    using maptype         = mapframe<typename goopax_struct_changetype<TT, xtype_int>::type>;
    using gpu_maptype     = mapframe<typename goopax_struct_changetype<TT, xtype_gpu>::type>;

    static constexpr unsigned int base_step = gettype<TT>::size / gettype<basetype>::size;

    protected:
    template<typename U>
    friend struct get_pointer_step;


    public:
    gpu_type& operator+=(const gpu_ssize_type& b)
    {
        gpu_ssize_type b2 = b * base_step;
        interface::goopax_4_gpu_type_op2(6, this->impl.get(), this->impl.get(), b2.impl.get());
        return *this;
    }
    gpu_type& operator-=(const gpu_ssize_type& b)
    {
        return *this += -b;
    }
    gpu_type operator+(gpu_ssize_type b) const
    {
        gpu_type ret = *this;
        ret += b;
        return ret;
    }
    friend gpu_type operator+(gpu_ssize_type a, gpu_type b)
    {
        return b + a;
    }

    gpu_type operator-(gpu_ssize_type b) const
    {
        return (*this) + (-b);
    }

    inline gpu_type& operator++()
    {
        /// increment
        return *this += 1;
    }
    inline gpu_type& operator--()
    {
        /// decrement
        return *this -= 1;
    }
    inline gpu_type operator++(int)
    {
        /// increment
        gpu_type ret = *this;
        ++*this;
        return ret;
    }
    inline gpu_type operator--(int)
    {
        /// decrement
        gpu_type ret = *this;
        --*this;
        return ret;
    }

    gpu_bool operator<(gpu_type b) const
    {
        gpu_bool ret;
        interface::goopax_4_gpu_type_op2(2, ret.impl.get(), this->impl.get(), b.impl.get());
        return ret;
    }

    gpu_bool operator>=(gpu_type b) const
    {
        gpu_bool ret;
        interface::goopax_4_gpu_type_op2(3, ret.impl.get(), this->impl.get(), b.impl.get());
        return ret;
    }

    gpu_bool operator<=(gpu_type b) const
    {
        return b >= *this;
    }
    gpu_bool operator>(gpu_type b) const
    {
        return b < *this;
    }

    reference operator*() const
    {
        value_type* ret = new value_type;
        interface::goopax_4_register_garbage(ret, garbage_deleter<value_type>);
        interface::goopax_4_sitem_set_memref(maptype::get_static_mf(), gpu_maptype::get_static_mf(), *this->impl, nullptr, gettype<basetype>::size, ret);
        return *ret;
    }

    reference operator[](gpu_ssize_type k) const
    {
        return *((*this) + k);
    }


    typename remove_reference<reference>::type* operator->() const
    {
        return &**this;
    }

    void gpu_break() const
    {
        /// Break from loop. Only valid on loop variables. Calling gpu_break() on non-loop variables results in an error.
        check_errno(interface::goopax_4E_loop_break(this->impl.get()));
    }
    void gpu_continue() const
    {
        /// Continue loop. Only valid on loop variables. Calling gpu_continue() on non-loop variables results in an error.
        check_errno(interface::goopax_4E_loop_continue(this->impl.get()));
    }

    template<typename EA = void, typename = typename enable_if<scope == memory::device, EA>::type>
    gpu_type(T* ptr)
    {
        this->impl.reset(interface::goopax_4_gpu_type_assign(&ptr, type_enum<basetype*>::value));
        /// conversion from CPU type
    }

    template<typename EA = void, typename EB = typename enable_if<std::is_const<T>::value, EA>::type>
    gpu_type(const gpu_type<typename remove_cv<T>::type*, scope>& b)
        : gpu_type()
    {
        interface::goopax_4_gpu_type_copyassign(this->impl.get(), b.impl.get());
    }

    gpu_type(const gpu_type& b)
        : gpu_type()
    {
        /// copy constructor
        *this = b;
    }
    gpu_type()
    {
        this->impl.reset(interface::goopax_4_gpu_type_uninit(type_enum<basetype*>::value, scope));
    }

    gpu_type& operator=(const gpu_type& b)
    {
        /// assignment operator
        interface::goopax_4_gpu_type_copyassign(this->impl.get(), b.impl.get());
        return *this;
    }

    template<typename EA = void, typename = typename enable_if<scope == memory::device, EA>::type>
    gpu_type& operator=(T* ptr)
    {
        return *this = gpu_type(ptr);
    }
};

template<typename T, memory::scope scope>
gpu_bool operator==(const gpu_type<T*, scope>& a, const gpu_type<T*, scope>& b)
{
    gpu_bool ret;
    interface::goopax_4_gpu_type_op2(0, ret.impl.get(), a.impl.get(), b.impl.get());
    return ret;
}
template<typename T, memory::scope scope>
gpu_bool operator!=(const gpu_type<T*, scope>& a, const gpu_type<T*, scope>& b)
{
    gpu_bool ret;
    interface::goopax_4_gpu_type_op2(1, ret.impl.get(), a.impl.get(), b.impl.get());
    return ret;
}

template<typename T, memory::scope scope>
gpu_bool operator==(const gpu_type<T*, scope>& a, const gpu_type<const T*, scope>& b)
{
    return (static_cast<gpu_type<const T*, scope>>(a) == b);
}
template<typename T, memory::scope scope>
gpu_bool operator!=(const gpu_type<T*, scope>& a, const gpu_type<const T*, scope>& b)
{
    return (static_cast<gpu_type<const T*, scope>>(a) != b);
}

template<typename T, memory::scope scope>
gpu_bool operator==(const gpu_type<const T*, scope>& a, const gpu_type<T*, scope>& b)
{
    return (a == static_cast<gpu_type<const T*, scope>>(b));
}
template<typename T, memory::scope scope>
gpu_bool operator!=(const gpu_type<const T*, scope>& a, const gpu_type<T*, scope>& b)
{
    return (a != static_cast<gpu_type<const T*, scope>>(b));
}

template<typename T, memory::scope scope>
gpu_bool operator==(const gpu_type<T*, scope>& a, std::nullptr_t null)
{
    return (a == gpu_type<T*, scope>(null));
}
template<typename T, memory::scope scope>
gpu_bool operator!=(const gpu_type<T*, scope>& a, std::nullptr_t null)
{
    return (a != gpu_type<T*, scope>(null));
}


template<typename T, memory::scope scope>
typename gpu_type<T*, scope>::gpu_ssize_type operator-(const gpu_type<T*, scope>& a, const gpu_type<T*, scope>& b)
{
    typename gpu_type<T*, scope>::gpu_ssize_type ret;
    interface::goopax_4_gpu_type_op2(18, ret.impl.get(), a.impl.get(), b.impl.get());
    return ret / gpu_type<T*, scope>::base_step;
}

template<typename T, memory::scope scope>
typename gpu_type<T*, scope>::gpu_ssize_type operator-(const gpu_type<const T*, scope>& a, const gpu_type<T*, scope>& b)
{
    return a - static_cast<gpu_type<const T*, scope>>(b);
}

template<typename T, memory::scope scope>
typename gpu_type<T*, scope>::gpu_ssize_type operator-(const gpu_type<T*, scope>& a, const gpu_type<const T*, scope>& b)
{
    return static_cast<gpu_type<const T*, scope>>(a) - b;
}

}
}
