// -*- C++ -*-
#pragma once

#include "pointer.h"
#include "struct.h"


namespace goopax {
namespace impl {
class svm_buffer_impl;

namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_svm_create(svm_buffer_impl**, goopax_device_impl*, size_t, void**, BUFFER_FLAGS flags) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_svm_map(svm_buffer_impl*, const void*, const void*, BUFFER_FLAGS) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_svm_unmap(svm_buffer_impl*, const void*) noexcept;
}
}

/// Shared Virtual Memory (called Unified Memory in CUDA)
template<typename T>
class svm_buffer
{
    static_assert(!is_checknan<T>::value, "sorry, debug types not yet allowed for svm memory");
    unique_ptr<svm_buffer_impl, interface::goopax_deleter> impl;
    T*                                                     ptr;
    size_t                                                 size_i;

    public:
    size_t size() const
    {
        return size_i;
    }

    T* data()
    {
        return ptr;
    }
    const T* data() const
    {
        return ptr;
    }

    T* begin()
    {
        return data();
    }
    const T* begin() const
    {
        return data();
    }
    T* end()
    {
        return begin() + size();
    }
    const T* end() const
    {
        return begin() + size();
    }

    T& operator[](size_t k)
    {
        return begin()[k];
    }
    const T& operator[](size_t k) const
    {
        return begin()[k];
    }

    /// prepare memory region for host access.
    /// \param begin start of mapped region
    /// \param end end of mapped region
    /// \param flags host access flags
    void map(const T* begin, const T* end, BUFFER_FLAGS flags = BUFFER_READ_WRITE) const
    {
        check_errno(interface::goopax_4E_svm_map(impl.get(), begin, end, flags));
    }

    /// prepare memory for host access.
    /// \param flags host access flags
    void map(BUFFER_FLAGS flags = BUFFER_READ_WRITE) const
    {
        map(this->begin(), this->end(), flags);
    }

    /// indicate that host access has completed
    /// \param begin begin pointer used when calling map
    void unmap(const T* begin) const
    {
        check_errno(interface::goopax_4E_svm_unmap(impl.get(), begin));
    }
    /// indicate that host access has completed
    void unmap() const
    {
        unmap(this->begin());
    }

    /// Construct svm memory region with shared address space between host and device.
    /// \param device device that should be able to access the memory
    /// \param size size in number of elements of type T
    /// \param flags gpu access flags
    svm_buffer(goopax_device device, size_t size, BUFFER_FLAGS flags = BUFFER_READ_WRITE)
        : size_i(size)
    {
        void* ptr;
        impl.reset(if_call(interface::goopax_4E_svm_create, device.get_impl(), size * sizeof(T), &ptr, flags));
        this->ptr = reinterpret<T*>(ptr);
    }
    svm_buffer(svm_buffer&&) = default;
    svm_buffer& operator=(svm_buffer&&) = default;

    ~svm_buffer()
    {
    }
};


}

using impl::svm_buffer;
}
